from yarl import URL

from ..base_functions.base_functions import log, logger, make_title_safe
from ..base_functions.data_classes import DomainItem
from ..client.client import Session


class XBunkrCrawler:
    def __init__(self, *, include_id=False):
        self.include_id = include_id

    async def fetch(self, session: Session, url: URL):
        domain_obj = DomainItem(url.host, {})

        if "media" in url.host:
            await domain_obj.add_to_album("Loose XBunkr Files", url, url)
            return domain_obj

        try:
            soup = await session.get_BS4(url)
            links = soup.select("a[class=image]")
            title = await make_title_safe(soup.select_one("h1[id=title]").text)
            title = title.strip()
            for link in links:
                link = URL(link.get('href'))
                await domain_obj.add_to_album(title, link, url)

        except Exception as e:
            logger.debug("Error encountered while handling %s", str(url), exc_info=True)
            await log("Error scraping " + str(url))
            logger.debug(e)

        await log("Finished scrape of " + str(url))

        return domain_obj



import atexit
import logging
import sqlite3
import os


class SQLHelper:
    def __init__(self, ignore_history, download_history):
        self.ignore_history = ignore_history
        self.download_history = download_history
        self.conn = None
        self.curs = None
        # Close the sql connection when the program exits
        atexit.register(self.exit_handler)

    async def sql_initialize(self):
        self.conn = sqlite3.connect(self.download_history)
        self.curs = self.conn.cursor()

        await self.create_tables()
        await self.pre_allocate()

        await self.check_columns()

    async def create_tables(self):
        create_table_query = """CREATE TABLE IF NOT EXISTS downloads (
                                            path TEXT,
                                            downloaded_filename TEXT,
                                            completed INTEGER NOT NULL,
                                            PRIMARY KEY (path)
                                        );"""
        create_temp_download_name_query = """CREATE TABLE IF NOT EXISTS downloads_temp (
                                                        downloaded_filename TEXT
                                                    );"""
        temp_truncate_query = """DELETE FROM downloads_temp;"""

        self.curs.execute(create_table_query)
        self.conn.commit()
        self.curs.execute(create_temp_download_name_query)
        self.conn.commit()
        self.curs.execute(temp_truncate_query)
        self.conn.commit()

    async def pre_allocate(self):
        pre_alloc = "CREATE TABLE IF NOT EXISTS t(x);"
        pre_alloc2 = "INSERT INTO t VALUES(zeroblob(50*1024*1024));"  # 50 mb
        drop_pre = "DROP TABLE t;"
        check_prealloc = "PRAGMA freelist_count;"

        self.curs.execute(check_prealloc)
        free = self.curs.fetchone()[0]
        if free <= 1024:
            self.curs.execute(pre_alloc)
            self.conn.commit()
            self.curs.execute(pre_alloc2)
            self.conn.commit()
            self.curs.execute(drop_pre)
            self.conn.commit()

    async def check_columns(self):
        self.curs.execute("""SELECT COUNT(*) AS CNTREC FROM pragma_table_info('downloads') WHERE 
                             name='path'""")
        sql_check = self.curs.fetchone()[0]
        if sql_check == (0):
            self.curs.execute("""DROP TABLE downloads""")
            self.conn.commit()
            create_table_query = """CREATE TABLE IF NOT EXISTS downloads (
                                    path TEXT,
                                    downloaded_filename TEXT,
                                    completed INTEGER NOT NULL,
                                    PRIMARY KEY (path)
                                );"""
            self.curs.execute(create_table_query)
            self.conn.commit()

    async def sql_check_existing(self, path):
        if self.ignore_history:
            return False
        self.curs.execute("""SELECT completed FROM downloads WHERE path = ?""", (path, ))
        sql_file_check = self.curs.fetchone()
        return sql_file_check and sql_file_check[0] == 1

    async def sql_insert_temp(self, downloaded_filename):
        self.curs.execute("""INSERT OR IGNORE INTO downloads_temp VALUES (?)""", (downloaded_filename,))
        self.conn.commit()

    async def sql_insert_file(self, path, downloaded_filename, completed):
        self.curs.execute("""INSERT OR IGNORE INTO downloads VALUES (?, ?, ?)""", (path, downloaded_filename, completed, ))
        self.conn.commit()

    async def sql_update_file(self, path, downloaded_filename, completed):
        self.curs.execute("""INSERT OR REPLACE INTO downloads VALUES (?, ?, ?)""", (path, downloaded_filename, completed, ))
        self.conn.commit()

    async def check_filename(self, filename):
        self.curs.execute("""SELECT EXISTS(SELECT 1 FROM downloads WHERE downloaded_filename = ?)""", (filename, ))
        sql_check = self.curs.fetchone()[0]
        return sql_check == 1

    async def get_download_filename(self, path):
        self.curs.execute("""SELECT downloaded_filename FROM downloads WHERE path = ?""", (path, ))
        filename = self.curs.fetchone()
        if filename:
            return filename[0]
        return None

    async def get_temp_names(self):
        self.curs.execute("SELECT downloaded_filename FROM downloads_temp;")
        filenames = self.curs.fetchall()
        filenames = list(sum(filenames, ()))
        return filenames

    def exit_handler(self):
        try:
            self.conn.commit()
            self.conn.close()
        except Exception as e:
            logging.debug(f"Failed to close sqlite database connection: {str(e)}")
        else:
            logging.debug("Successfully closed sqlite database connection")



from yarl import URL

from ..base_functions.base_functions import log, logger, make_title_safe, check_direct
from ..base_functions.data_classes import DomainItem
from ..client.client import Session


class HGameCGCrawler:
    async def fetch(self, session: Session, url: URL):
        domain_obj = DomainItem(url.host, {})

        await log("Starting scrape of " + str(url))
        results = await self.get_album(session, url)
        for result in results:
            await domain_obj.add_to_album(result['title'], result['url'], result['referral'])
        await log("Finished scrape of " + str(url))
        return domain_obj

    async def get_album(self, session: Session, url: URL):
        results = []
        try:
            soup = await session.get_BS4(url)
            title = await make_title_safe(soup.select_one("div[class=navbar] h1").get_text())
            results = []

            images = soup.select("div[class=image] a")
            for image in images:
                image = image.get('href')
                image = URL("https://" + url.host + image)
                link = await self.get_image(session, image)
                results.append({'url': link, 'title': title, 'referral': url, 'cookies': ''})

            next_page = soup.find("a", text="Next Page")
            if next_page is not None:
                next_page = next_page.get('href')
                if next_page is not None:
                    next_page = URL("https://" + url.host + next_page)
                    results.extend(result for result in await self.get_album(session, next_page))

        except Exception as e:
            logger.debug("Error encountered while handling %s", str(url), exc_info=True)
            await log("Error scraping " + str(url))
            logger.debug(e)

        return results

    async def get_image(self, session: Session, url: URL):
        try:
            soup = await session.get_BS4(url)
            image = soup.select_one("div[class=hgamecgimage] img")
            image = URL(image.get('src'))
            return image
        except:
            logger.error(f"Unable to get image from {str(url)}")



import asyncio

from yarl import URL

from ..base_functions.base_functions import log, logger, make_title_safe
from ..base_functions.data_classes import DomainItem
from ..client.client import Session


class KemonoCrawler:
    def __init__(self, *, include_id=False, scraping_mapper, separate_posts=False):
        self.include_id = include_id
        self.scraping_mapper = scraping_mapper
        self.separate_posts = separate_posts

    async def fetch(self, session: Session, url: URL):
        await log("Starting scrape of " + str(url))
        domain_obj = DomainItem('kemono.party', {})
        results = []

        if "thumbnail" in url.parts:
            parts = [x for x in url.parts if x not in ("thumbnail", "/")]
            link = URL("https://kemono.party/" + "/".join(parts))
            await domain_obj.add_to_album("Loose Kemono.Party Files", link, link)
        elif "data" in url.parts:
            await domain_obj.add_to_album("Loose Kemono.Party Files", url, url)
        elif "post" in url.parts:
            results = await self.parse_post(session, url, None)
            for result in results:
                await domain_obj.add_to_album(result[0], result[1], result[2])
        else:
            results.extend(await self.parse_profile(session, url))

        for result in results:
            await domain_obj.add_to_album(result[0], result[1], result[2])

        await log("Finished scrape of " + str(url))
        return domain_obj

    async def parse_profile(self, session: Session, url: URL):
        try:
            soup = await session.get_BS4(url)
            title = soup.select_one("span[itemprop=name]").get_text()
            title = title + " (Kemono.party)"
            results = []

            posts = soup.select('article[class="post-card post-card--preview"] a')
            for post in posts:
                path = post.get('href')
                if path:
                    post_link = URL("https://kemono.party" + path)
                    results.extend(await self.parse_post(session, post_link, title))

            next_page = soup.select_one('a[class=next]')
            if next_page:
                next_page = next_page.get('href')
                if next_page:
                    results.extend(await self.parse_profile(session, URL("https://kemono.party" + next_page)))
            return results

        except Exception as e:
            logger.debug("Error encountered while handling %s", str(url), exc_info=True)
            await log("Error scraping " + str(url))
            logger.debug(e)
            return []

    async def parse_post(self, session: Session, url: URL, title=None):
        try:
            soup = await session.get_BS4(url)
            results = []

            if self.separate_posts:
                if title:
                    title = title + '/' + await make_title_safe(soup.select_one("h1[class=post__title]").text.replace('\n', '').replace("..", ""))
                else:
                    title = await make_title_safe(soup.select_one("h1[class=post__title]").text.replace('\n', '').replace("..", ""))
            else:
                if not title:
                    title = await make_title_safe(soup.select_one("h1[class=post__title]").text.replace('\n', '').replace("..", ""))

            images = soup.select('a[class="fileThumb"]')
            for image in images:
                image_link = URL("https://kemono.party" + image.get('href'))
                results.append([title, image_link, url])

            downloads = soup.select('a[class=post__attachment-link]')
            for download in downloads:
                download_link = URL("https://kemono.party" + download.get('href'))
                results.append([title, download_link, url])

            text_content = soup.select("div[class=post__content] a")
            tasks = []
            for content in text_content:
                link = URL(content.get('href'))
                tasks.append(self.scraping_mapper.map_url(link, title))
            await asyncio.gather(*tasks)

            return results

        except Exception as e:
            logger.debug("Error encountered while handling %s", str(url), exc_info=True)
            await log("Error scraping " + str(url))
            logger.debug(e)
            return []



from typing import Dict

import aiofiles
import logging
import os
import re
from pathlib import Path

import yaml
from colorama import Fore, Style
from yarl import URL

from .config_schema import config_default, files_args, authentication_args, jdownloader_args, runtime_args

"""This file contains generic information and functions that are used around the program"""

FILE_FORMATS = {
    'Images': {
        '.jpg', '.jpeg', '.png', '.gif',
        '.webp', '.jpe', '.svg', '.jfif',
        '.tif', '.tiff', '.jif',
    },
    'Videos': {
        '.mpeg', '.avchd', '.webm', '.mpv',
        '.swf', '.avi', '.m4p', '.wmv',
        '.mp2', '.m4v', '.qt', '.mpe',
        '.mp4', '.flv', '.mov', '.mpg',
        '.ogg', '.mkv', '.mts', '.ts',
    },
    'Audio': {
        '.mp3', '.flac', '.wav', '.m4a',
    },
    'Other': {
        '.json', '.torrent', '.zip', '.rar',
        '.7z', '.torrent', '.psd', '.pdf',
    }
}

MAX_FILENAME_LENGTH = 100

logger = logging.getLogger(__name__)


class FailureException(Exception):
    """Basic failure exception I can throw to force a retry."""

    def __init__(self, code, message="Something went wrong"):
        self.code = code
        self.message = message
        super().__init__(self.message)
        super().__init__(self.code)


async def sanitize(name: str) -> str:
    return re.sub(r'[<>:"/\\|?*\']', "", name).strip()


async def log(text, style=Fore.WHITE) -> None:
    logger.debug(text)
    print(style + str(text) + Style.RESET_ALL)


async def clear() -> None:
    """Clears the terminal screen"""
    os.system('cls' if os.name == 'nt' else 'clear')


async def make_title_safe(title: str):
    title = re.sub(r'[\\*?:"<>|./]', "-", title)
    return title


async def purge_dir(dirname, in_place=True):
    deleted = []
    dir_tree = list(os.walk(dirname, topdown=False))

    for tree_element in dir_tree:
        sub_dir = tree_element[0]
        dir_count = len(os.listdir(sub_dir))
        if dir_count == 0:  # Helps with readability and i've had issues with it deleting non-empty dirs
            deleted.append(sub_dir)

    if in_place:
        list(map(os.rmdir, deleted))


async def regex_links(urls) -> list:
    all_links = [x.group().replace(".md.", ".") for x in re.finditer(
        r"(?:http.*?)(?=($|\n|\r\n|\r|\s|\"|\[/URL]|]\[|\[/img]))", urls)]
    yarl_links = []
    for link in all_links:
        yarl_links.append(URL(link))
    return yarl_links


async def cyberdrop_parse(url: URL) -> URL:
    mapping_direct = [r'img-...cyberdrop...',
                      r'f.cyberdrop...', r'fs-...cyberdrop...']
    url = str(url)
    for mapping in mapping_direct:
        url = re.sub(mapping, 'cyberdrop.to', url)
    return URL(url)


async def write_last_post_file(file: Path, url: str):
    async with aiofiles.open(file, mode='a') as f:
        await f.write(url + '\n')
    return


async def check_direct(url: URL):
    mapping_direct = [r'i.pixl.is', r'i..pixl.is', r'img-...cyberdrop...', r'f.cyberdrop...',
                      r'fs-...cyberdrop...', r'i.bunkr...', r'i..bunkr...', r'i...bunkr...', r'media-files.bunkr...',
                      r'media-files..bunkr...', r'cdn.bunkr...', r'cdn..bunkr...', r'cdn...bunkr...',
                      r'jpg.church/images/...', r'simp..jpg.church', r's..putmega.com', r's..putme.ga',]
    return any(re.search(domain, url.host) for domain in mapping_direct)


async def create_config(config: Path, passed_args=None, remake=None):
    if config.is_file() and not remake:
        logging.debug("Validating Config")
        await validate_config(config)
        return

    logging.debug("Creating Config File")
    config_data = config_default
    if passed_args:
        for arg in authentication_args:
            if arg in passed_args.keys():
                config_data[0]["Configuration"]["Authentication"][arg] = passed_args[arg]
        for arg in files_args:
            if arg in passed_args.keys():
                config_data[0]["Configuration"]["Files"][arg] = str(passed_args[arg])
        for arg in jdownloader_args:
            if arg in passed_args.keys():
                config_data[0]["Configuration"]["JDownloader"][arg] = passed_args[arg]
        for arg in runtime_args:
            if arg in passed_args.keys():
                config_data[0]["Configuration"]["Runtime"][arg] = passed_args[arg]

    with open(config, 'w') as yamlfile:
        yaml.dump(config_data, yamlfile)
    return


async def validate_config(config: Path):
    with open(config, "r") as yamlfile:
        data = yaml.load(yamlfile, Loader=yaml.FullLoader)
    data = data[0]["Configuration"]
    recreate = 0
    try:
        if not set(authentication_args).issubset(set(data['Authentication'].keys())):
            recreate = 1
        if not set(files_args).issubset(set(data['Files'].keys())):
            recreate = 1
        if not set(jdownloader_args).issubset(set(data['JDownloader'].keys())):
            recreate = 1
        if not set(runtime_args).issubset(set(data['Runtime'].keys())):
            recreate = 1

        if recreate:
            config.unlink()
            logging.debug("Recreating Config")

            args = {}
            args_list = [data['Authentication'], data['Files'], data['JDownloader'], data['Runtime']]
            for dic in args_list:
                args.update(dic)
            await create_config(config, args, True)

    except KeyError:
        config.unlink()
        await log("Config was malformed, recreating.")
        await create_config(config, None, True)


async def run_args(config: Path, cmd_arg: Dict):
    with open(config, "r") as yamlfile:
        data = yaml.load(yamlfile, Loader=yaml.FullLoader)
    data = data[0]["Configuration"]
    if data['Apply_Config']:
        return data

    logging.debug("Gathering Args")
    config_data = config_default[0]["Configuration"]
    for arg in authentication_args:
        if arg in cmd_arg.keys():
            config_data["Authentication"][arg] = cmd_arg[arg]
    for arg in files_args:
        if arg in cmd_arg.keys():
            config_data["Files"][arg] = cmd_arg[arg]
    for arg in jdownloader_args:
        if arg in cmd_arg.keys():
            config_data["JDownloader"][arg] = cmd_arg[arg]
    for arg in runtime_args:
        if arg in cmd_arg.keys():
            config_data["Runtime"][arg] = cmd_arg[arg]
    return config_data



import argparse
import multiprocessing
from argparse import Namespace
import asyncio
import logging
from pathlib import Path
from functools import wraps
from typing import Dict

from colorama import Fore
from yarl import URL

from . import __version__ as VERSION
from cyberdrop_dl.base_functions.base_functions import clear, create_config, log, logger, purge_dir, regex_links, \
    run_args
from cyberdrop_dl.base_functions.data_classes import AuthData, SkipData
from cyberdrop_dl.base_functions.sql_helper import SQLHelper
from cyberdrop_dl.client.client import Client
from cyberdrop_dl.client.downloaders import get_downloaders
from cyberdrop_dl.scraper.scraper import scrape


def parse_args():
    parser = argparse.ArgumentParser(description="Bulk downloader for multiple file hosts")
    parser.add_argument("-V", "--version", action="version", version="%(prog)s " + VERSION)
    parser.add_argument("-i", "--input-file", type=Path, help="file containing links to download", default="URLs.txt")
    parser.add_argument("-o", "--output-folder", type=Path, help="folder to download files to", default="Downloads")
    parser.add_argument("--log-file", type=Path, help="log file to write to", default="downloader.log")
    parser.add_argument("--config-file", type=Path, help="config file to read arguments from", default="config.yaml")
    parser.add_argument("--db-file", type=Path, help="history database file to write to", default="download_history.sqlite")
    parser.add_argument("--threads", type=int, help="number of threads to use (0 = max)", default=0)
    parser.add_argument("--attempts", type=int, help="number of attempts to download each file", default=10)
    parser.add_argument("--connection-timeout", type=int, help="number of seconds to wait attempting to connect to a URL during the downloading phase", default=15)
    parser.add_argument("--disable-attempt-limit", help="disables the attempt limitation", action="store_true")
    parser.add_argument("--include-id", help="include the ID in the download folder name", action="store_true")
    parser.add_argument("--exclude-videos", help="skip downloading of video files", action="store_true")
    parser.add_argument("--exclude-images", help="skip downloading of image files", action="store_true")
    parser.add_argument("--exclude-audio", help="skip downloading of audio files", action="store_true")
    parser.add_argument("--exclude-other", help="skip downloading of images", action="store_true")
    parser.add_argument("--ignore-history", help="This ignores previous download history", action="store_true")
    parser.add_argument("--output-last-forum-post", help="Separates forum scraping into folders by post number", action="store_true")
    parser.add_argument("--proxy", help="HTTP/HTTPS proxy used for downloading, format [protocal]://[ip]:[port]", default=None)
    parser.add_argument("--separate-posts", help="Separates forum scraping into folders by post number", action="store_true")
    parser.add_argument("--mark-downloaded", help="Sets the scraped files as downloaded without downloading", action="store_true")
    parser.add_argument("--xbunker-username", type=str, help="username to login to xbunker", default=None)
    parser.add_argument("--xbunker-password", type=str, help="password to login to xbunker", default=None)
    parser.add_argument("--socialmediagirls-username", type=str, help="username to login to socialmediagirls", default=None)
    parser.add_argument("--socialmediagirls-password", type=str, help="password to login to socialmediagirls", default=None)
    parser.add_argument("--simpcity-username", type=str, help="username to login to simpcity", default=None)
    parser.add_argument("--simpcity-password", type=str, help="password to login to simpcity", default=None)
    parser.add_argument("--jdownloader-enable", help="enables sending unsupported URLs to a running jdownloader2 instance to download", action="store_true")
    parser.add_argument("--jdownloader-username", type=str, help="username to login to jdownloader", default=None)
    parser.add_argument("--jdownloader-password", type=str, help="password to login to jdownloader", default=None)
    parser.add_argument("--jdownloader-device", type=str, help="device name to login to for jdownloader", default=None)
    parser.add_argument("--skip", dest="skip_hosts", choices=SkipData.supported_hosts, help="This removes host links from downloads", action="append", default=[])
    parser.add_argument("--ratelimit", type=int, help="this will add a ratelimiter to requests made in the program during scraping, the number you provide is in requests/seconds", default=50)
    parser.add_argument("--throttle", type=int, help="This is a throttle between requests during the downloading phase, the number is in seconds", default=0.5)
    parser.add_argument("links", metavar="link", nargs="*", help="link to content to download (passing multiple links is supported)", default=[])
    args = parser.parse_args()
    return args


async def handle_args(args: argparse.Namespace):
    print_args = Namespace(**vars(args)).__dict__
    print_args['xbunker_password'] = '!REDACTED!'
    print_args['socialmediagirls_password'] = '!REDACTED!'
    print_args['simpcity_password'] = '!REDACTED!'
    print_args['jdownloader_password'] = '!REDACTED!'

    logging.debug("Creating Config")
    cmd_args = Namespace(**vars(args)).__dict__
    await create_config(args.config_file, print_args)

    await log("Getting Run Args")
    use_args = await run_args(args.config_file, cmd_args)

    auth_args = use_args['Authentication']
    auth_args_print = auth_args.copy()
    auth_args_print['xbunker_password'] = '!REDACTED!'
    auth_args_print['socialmediagirls_password'] = '!REDACTED!'
    auth_args_print['simpcity_password'] = '!REDACTED!'

    file_args = use_args['Files']
    for key, value in file_args.items():
        file_args[key] = Path(value)

    jdownloader_args = use_args['JDownloader']
    jdownloader_args_print = jdownloader_args
    jdownloader_args_print['jdownloader_password'] = '!REDACTED!'

    runtime_args = use_args['Runtime']

    logging.debug(f"Starting Cyberdrop-DL")
    logging.debug(f"Using authorization arguments: {auth_args_print}")
    logging.debug(f"Using file arguments: {file_args}")
    logging.debug(f"Using jdownloader arguments: {jdownloader_args}")
    logging.debug(f"Using runtime arguments: {runtime_args}")

    return auth_args, file_args, jdownloader_args, runtime_args


async def download_all(auth_args: Dict, file_args: Dict, jdownloader_args: Dict, runtime_args: Dict, links: list,
                       client: Client, SQL_helper: SQLHelper, threads: int):
    xbunker_auth = AuthData(auth_args['xbunker_username'], auth_args['xbunker_password'])
    socialmediagirls_auth = AuthData(auth_args['socialmediagirls_username'], auth_args['socialmediagirls_password'])
    simpcity_auth = AuthData(auth_args['simpcity_username'], auth_args['simpcity_password'])
    jdownloader_auth = AuthData(jdownloader_args['jdownloader_username'], jdownloader_args['jdownloader_password'])

    skip_data = SkipData(runtime_args['skip_hosts'])
    excludes = {'videos': runtime_args['exclude_videos'], 'images': runtime_args['exclude_images'],
                'audio': runtime_args['exclude_audio'], 'other': runtime_args['exclude_other']}

    content_object = await scrape(urls=links, client=client, file_args=file_args, jdownloader_args=jdownloader_args,
                                  runtime_args=runtime_args, jdownloader_auth=jdownloader_auth,
                                  simpcity_auth=simpcity_auth, socialmediagirls_auth=socialmediagirls_auth,
                                  xbunker_auth=xbunker_auth, skip_data=skip_data)

    if await content_object.is_empty():
        logging.error('ValueError No links')
        await log("No links found duing scraping, check passwords or that the urls are accessible", Fore.RED)
        await log("This program does not currently support password protected albums.", Fore.RED)
        exit(0)
    await clear()

    downloaders = await get_downloaders(content_object, excludes=excludes, SQL_helper=SQL_helper, client=client,
                                        max_workers=threads, file_args=file_args, runtime_args=runtime_args)

    for downloader in downloaders:
        await downloader.download_content()


async def director(args: argparse.Namespace):
    await clear()
    await log(f"We are running version {VERSION} of Cyberdrop Downloader", Fore.WHITE)

    auth_args, file_args, jdownloader_args, runtime_args = await handle_args(args)

    input_file = file_args['input_file']
    if not input_file.is_file():
        input_file.touch()
        await log(f"{input_file} created. Populate it and retry.")
        exit(1)

    client = Client(runtime_args['ratelimit'], runtime_args['throttle'])
    SQL_helper = SQLHelper(runtime_args['ignore_history'], file_args['db_file'])
    await SQL_helper.sql_initialize()

    threads = runtime_args['threads'] if runtime_args['threads'] != 0 else multiprocessing.cpu_count()

    links = args.links
    links = list(map(URL, links))

    with open(input_file, "r", encoding="utf8") as f:
        links += await regex_links(f.read())
    links = list(filter(None, links))

    if not links:
        await log("No links found, check the URL.txt\nIf the link works in your web browser, "
                  "please open an issue ticket with me.", Fore.RED)

    if runtime_args['output_last_forum_post']:
        output_url_file = file_args['output_last_forum_post_file']
        if output_url_file.exists():
            output_url_file.unlink()
            output_url_file.touch()

    await download_all(auth_args=auth_args, file_args=file_args, jdownloader_args=jdownloader_args,
                       runtime_args=runtime_args, links=links, client=client, SQL_helper=SQL_helper, threads=threads)

    logger.debug("Finished")

    partial_downloads = [str(f) for f in file_args['output_folder'].rglob("*.part") if f.is_file()]
    temp_downloads_check = [str(f) for f in await SQL_helper.get_temp_names() if Path(f).is_file()]

    await log('Purging empty directories')
    await purge_dir(file_args['output_folder'])

    await log('Finished downloading. Enjoy :)')
    if partial_downloads:
        await log('There are partial downloads in the downloads folder.')
    if temp_downloads_check:
        await log('There are partial downloads from this run, please re-run the program.')


def main(args=None):
    if not args:
        args = parse_args()
    logging.basicConfig(
        filename=args.log_file,
        level=logging.DEBUG,
        format="%(asctime)s:%(levelname)s:%(module)s:%(filename)s:%(lineno)d:%(message)s",
        filemode="w"
    )

    try:
        loop = asyncio.get_event_loop()
        loop.run_until_complete(director(args))
        loop.run_until_complete(asyncio.sleep(5))
    except RuntimeError:
        pass


if __name__ == '__main__':
    print("""STOP! If you're just trying to download files, check the README.md file for instructions.
    If you're developing this project, use start.py instead.""")
    exit()
